import pygame, sys, os
from pygame.locals import *

EID_PLAYER=0
EID_SLASH=1
EID_FLASH=2

ETYPE_FRIENDLY=0
ETYPE_ENEMY=1

input_keys=[]

lightness=True

solids=[]
solids1=[]
solids2=[]
entities=[]
animations=[]

# enemy arrays
werepollies=[]

# ------

def zrect(x, y, w, h):
	return pygame.Rect(x*3, y*3, w*3, h*3)

def pgrect(x, y, w, h):
	return pygame.Rect(x, y, w, h)

def loadimage (filename):
	return pygame.image.load(os.path.join("data", filename))

# ------

##
class Template:
	def __init__(self, rect1, rect2, solid):
		self.brect=rect1
		self.crect=rect2
		self.solid=solid

##
class Block:
	def __init__(self, sheet1, sheet2, srect, crect, x, y, solid, darkside):
		self.image1=sheet1.subsurface(srect)
		self.image2=sheet2.subsurface(srect)
		if solid:
			if not darkside:
				solids1.append(crect.move(x, y))
			else:
				solids2.append(crect.move(x, y))
		self.x=x*3
		self.y=y*3
##
class Animation:
	def __init__(self, sheet, frames, speed):
		self.length=0
		self.frames=[]
		for frame in frames:
			self.frames.append(sheet.subsurface(frame))
			self.length+=1
#		self.frames=frames
		self.speed=speed
##
class Entity:
	def __init__(self, crect, x, y, etype):
		self.alive=True
		self.etype=etype
		self.rect=crect
		self.cbox=self.rect.move(x, y)
		self.x=x
		self.y=y
		self.fx=x*256
		self.fy=y*256
		self.dx=0
		self.dy=0
		self.gravity=0
		self.jumping=0
		self.flip=0
		self.animation=None
		self.apos=0
		self.adelay=0
		self.edgeleft=False
		self.edgeright=False
		self.health=0
		self.invincible=0
		self.hurting=False
		self.flag1=False
		self.flag2=False
		self.counter1=0
		self.counter2=0
	def attacked(self, abox, amount):
		if self.invincible==0 and self.cbox.colliderect(abox):
			self.health-=amount
			if self.health<0:
				self.health=0
			self.hurting=True
			if self.etype==ETYPE_FRIENDLY: # this means it's the player... ugly :/
				if abox.x+abox.width/2>self.cbox.x+self.cbox.width/2:
					self.flip=False
					self.dx=-256
				else:
					self.flip=True
					self.dx=256
				self.dy=-512
				self.invincible=60
				self.jumping=30
			else:
				self.invincible=25 # enemies just pause for a bit when hurt (implemented in their update function)
			entities[EID_FLASH].alive=True
			entities[EID_FLASH].apos=0
			entities[EID_FLASH].adelay=0
			entities[EID_FLASH].counter1=11
			entities[EID_FLASH].fy=((self.cbox.y+self.cbox.height/2+abox.y+abox.height/2)/2-8)*256
			entities[EID_FLASH].fx=((self.cbox.x+self.cbox.width/2+abox.x+abox.width/2)/2-8)*256
			entities[EID_FLASH].setanimation(7)
	def animate(self):
		if not self.animation==None:
			self.adelay+=1
			if self.adelay>=self.animation.speed:
				self.apos+=1
				self.adelay=0
				if self.apos>=self.animation.length:
					self.apos=0
	def setanimation(self, index):
		if not self.animation==animations[index]:
			self.animation=animations[index]
			self.apos=0
			self.adelay=0
	def getsprite(self):
		return self.animation.frames[self.apos]
	def update(self):
		self.dy+=self.gravity
		if self.dy>512:
			self.dy=512
		self.fx+=self.dx
		if self.dy>-512:
			self.fy+=self.dy
		else:
			self.fy-=512
		footrect=pgrect(self.fx/256+self.rect.x, self.fy/256+30, self.rect.width, 2)
		if self.dy>0:
			if footrect.collidelist(solids)>-1:
				self.fy=(self.fy-256)/512*512
				self.dy=0
				self.jumping=0
			else:
				self.jumping+=1
		self.edgeleft=False
		self.edgeright=False
		footrect=pgrect(self.fx/256+self.rect.x-3, self.fy/256+32, 3, 3)
		if footrect.collidelist(solids)==-1:
			self.edgeleft=True
		footrect=pgrect(self.fx/256+self.rect.x+self.rect.width, self.fy/256+32, 3, 3)
		if footrect.collidelist(solids)==-1:
			self.edgeright=True
		self.x=self.fx/256
		self.y=self.fy/256
		self.animate()
		self.cbox=self.rect.move(self.x, self.y)
		if self.invincible>0:
			self.invincible-=1
			if self.jumping<6:
				self.hurting=False
##
class Werepolly:
	def __init__(self, entity):
		self.entity=entity
		self.entity.dx=128

##
class Level:
	def __init__(self):
		self.width=640
		self.height=480

# ------

##
def update_player():
	global lightness
	move=False
	player=entities[EID_PLAYER]
	if player.counter1==0 and not player.hurting: # if not attacking and not hurting
		if player.counter2==0 or player.counter2>=120:
			if input_keys[0]: # move right
				player.flip=False
				move=True
				player.dx+=64
				if player.dx>256:
					player.dx=256
			if input_keys[1]: # move left
				player.flip=True
				move=True
				player.dx-=64
				if player.dx<-256:
					player.dx=-256
			if input_keys[3]: # attack
				if player.jumping<6 and not player.flag2:
					player.flag2=True
					player.counter1=29
					player.setanimation(4)
		if input_keys[4]: # vanish
			if player.jumping<6 and player.counter2<120:
				player.counter2+=1
				player.setanimation(5)
				slash=entities[EID_SLASH]
				slash.setanimation(10)
				slash.dx=0
				if not player.flip:
					slash.fx=player.fx+20*256
				else:
					slash.fx=player.fx+5*256
				slash.fy=player.fy-14*256
				slash.alive=True
				if player.counter2>=120:
					lightness=not lightness
					entities[EID_SLASH].alive=False
		else:
			if player.counter2>0 and player.counter2<120:
				player.counter2=0
				entities[EID_SLASH].alive=False
	if player.counter2>=120:
		player.counter2+=1
		if player.counter2>180:
			player.counter2=0

	if not move: # slow down if not moving
		if player.invincible==0 or player.jumping<6:
			if player.counter1==0 and (player.counter2==0 or player.counter2>=120):
				player.setanimation(0)
			if player.dx>0:
				player.dx-=64
			if player.dx<0:
				player.dx+=64
	else:
		if player.counter1==0:
			player.setanimation(1)

	if player.hurting:
		player.setanimation(3)
	else:
		if player.jumping>6:
			player.setanimation(2)
	if input_keys[2] and player.jumping<21 and not player.flag1 and player.counter1==0: # jump
		if player.jumping<6:
			player.jumping=6
		player.jumping+=1
		player.dy=-512
	if player.jumping>6 and not input_keys[2]: # disable jumping until landed
		player.jumping=22
	if player.jumping>=22:
		player.flag1=True
	if not input_keys[2]: # released jump button
		player.flag1=False
	if not input_keys[3]: # released attack button
		player.flag2=False

	if player.counter1>0:
		player.counter1-=1
		slash=entities[EID_SLASH]
		if player.counter1==19: # spawn sword slash
			slash.fy=player.fy+14*256
			if not player.flip:
				slash.fx=player.fx+22*256
				swordbox=pygame.Rect(player.x+22, player.y+14, 16, 14)
				slash.dx=256
				slash.flip=False
			else:
				slash.fx=player.fx+2*256
				swordbox=pygame.Rect(player.x-6, player.y+14, 16, 14)
				slash.dx=-256
				slash.flip=True
			slash.apos=0
			slash.adelay=0
			slash.setanimation(6)
			slash.alive=True
			for victim in entities:
				if victim.etype==ETYPE_ENEMY:
					victim.attacked(swordbox, 1)
		if player.counter1==10: # remove sword slash
			slash.alive=False
			slash.dx=0

##
def update_werepolly(specimen):
	polly=specimen.entity
	if polly.invincible>0:
		if polly.health==0:
			polly.setanimation(9)
		polly.dx=0
		if polly.invincible==1: # end of hurting, turn towards player
			if polly.health==0:
				polly.alive=False
			if entities[EID_PLAYER].x>polly.x:
				polly.dx=128
				polly.flip=False
			else:
				polly.dx=-128
				polly.flip=True
	else:
		if polly.jumping<6:
			if polly.edgeright and polly.dx>0:
				polly.dx=-128
				polly.flip=True
			if polly.edgeleft and polly.dx<0:
				polly.dx=128
				polly.flip=False
	entities[EID_PLAYER].attacked(polly.cbox, 2)

##
def update_hitflash():
	flash=entities[EID_FLASH]
	flash.counter1-=1
	if flash.counter1<=0:
		flash.alive=False

# ------

def main():
	pygame.display.init()
	screen=pygame.display.set_mode([640,480], DOUBLEBUF) # HWSURFACE|FULLSCREEN|DOUBLEBUF

	bsheet1=loadimage("blocksheet1.tga")
	bsheet1.set_colorkey([160,160,255], 0)
	bsheet1=pygame.transform.scale(bsheet1, (bsheet1.get_width()*3, bsheet1.get_height()*3))
	bsheet2=loadimage("blocksheet2.tga")
	bsheet2.set_colorkey([160,160,255], 0)
	bsheet2=pygame.transform.scale(bsheet2, (bsheet2.get_width()*3, bsheet2.get_height()*3))
	ssheet=loadimage("spritesheet.tga")
	ssheet.set_colorkey([160,160,255], 0)
	ssheet=pygame.transform.scale(ssheet, (ssheet.get_width()*3, ssheet.get_height()*3))
	title=loadimage("title.tga")
	title=pygame.transform.scale(title, (title.get_width()*3, title.get_height()*3))

	ssheet=ssheet.convert()
	bsheet1=bsheet1.convert()
	bsheet2=bsheet2.convert()

	animations.append(Animation(ssheet, [zrect(0*32,32,32,32)], 10)) # 0 - player idle
	animations.append(Animation(ssheet, [zrect(1*32,32,32,32), zrect(2*32,32,32,32)], 15)) # 1 - player run
	animations.append(Animation(ssheet, [zrect(3*32,32,32,32)], 10)) # 2 - player jump
	animations.append(Animation(ssheet, [zrect(4*32,32,32,32)], 10)) # 3 - player hurt
	animations.append(Animation(ssheet, [zrect(5*32,32,32,32), zrect(6*32,32,32,32), zrect(6*32,32,32,32)], 10)) # 4 - player attack
	animations.append(Animation(ssheet, [zrect(7*32,32,32,32)], 10)) # 5 - player vanish
	animations.append(Animation(ssheet, [zrect(0*8,0,8,15), zrect(0,3*32-1,1,1), zrect(1*8,0,8,15)], 4)) # 6 - sword slash
	animations.append(Animation(ssheet, [zrect(32+0*16,0,16,16), zrect(32+1*16,0,16,16), zrect(32+2*16,0,16,16), zrect(32+3*16,0,16,16)], 3)) # 7 - hit flash
	animations.append(Animation(ssheet, [zrect(0*32,2*32,32,32), zrect(1*32,2*32,32,32)], 15)) # 8 - werepolly walk
	animations.append(Animation(ssheet, [zrect(2*32,2*32,32,32), zrect(0,3*32-1,1,1)], 4)) # 9 - werepolly death
	animations.append(Animation(ssheet, [zrect(100,0,7,15), zrect(90,0,7,15), zrect(107,0,7,15), zrect(90,0,7,15), zrect(114,0,7,15), zrect(90,0,7,15)], 3)) # 10 - vanishing sword effect
	
	hud_health1=ssheet.subsurface(zrect(0,16,5,11))
	hud_health2=ssheet.subsurface(zrect(6,16,5,11))

	templates=[]
	templates.append(Template(zrect(0,0,66,4), pgrect(0, 0, 66, 4), True))
	templates.append(Template(zrect(68,0,8,3), pgrect(0, 0, 8, 4), True))
	templates.append(Template(zrect(83,26,17,6), pgrect(0, 0, 17, 4), True))
	templates.append(Template(zrect(0,67,32,6), pgrect(0, 0, 32, 4), True))
	templates.append(Template(zrect(37,67,17,3), pgrect(0, 0, 17, 4), True))
	templates.append(Template(zrect(59,67,37,2), pgrect(0, 0, 37, 4), True))

	templates.append(Template(zrect(160,0,9,50), pgrect(0, 0, 0, 0), False))
	templates.append(Template(zrect(170,0,24,11), pgrect(0, 0, 0, 0), False))
	templates.append(Template(zrect(195,0,8,10), pgrect(0, 0, 0, 0), False))
	templates.append(Template(zrect(205,0,3,10), pgrect(0, 0, 0, 0), False))
	templates.append(Template(zrect(182,20,3,21), pgrect(0, 0, 0, 0), False))
	templates.append(Template(zrect(193,20,21,21), pgrect(0, 0, 0, 0), False))
	templates.append(Template(zrect(219,20,21,21), pgrect(0, 0, 0, 0), False))

	templates.append(Template(zrect(3,5,12,16), pgrect(0, 0, 0, 0), False))
	templates.append(Template(zrect(51,5,12,16), pgrect(0, 0, 0, 0), False))
	templates.append(Template(zrect(23,5,18,16), pgrect(0, 0, 0, 0), False))
	templates.append(Template(zrect(68,10,4,16), pgrect(0, 0, 0, 0), False))
	templates.append(Template(zrect(76,7,1,24), pgrect(0, 0, 0, 0), False))
	templates.append(Template(zrect(91,10,27,12), pgrect(0, 0, 0, 0), False))
	templates.append(Template(zrect(92,3,5,6), pgrect(0, 0, 0, 0), False))
	templates.append(Template(zrect(100,3,5,6), pgrect(0, 0, 0, 0), False))
	templates.append(Template(zrect(109,3,5,6), pgrect(0, 0, 0, 0), False))
	templates.append(Template(zrect(116,3,2,6), pgrect(0, 0, 0, 0), False))
	templates.append(Template(zrect(121,3,1,6), pgrect(0, 0, 0, 0), False))
	templates.append(Template(zrect(120,10,8,13), pgrect(0, 0, 0, 0), False))
	templates.append(Template(zrect(130,0,7,20), pgrect(0, 0, 0, 0), False))
	templates.append(Template(zrect(0,34,39,24), pgrect(0, 0, 0, 0), False))

	blocks1=[]
	blocks2=[]

	# player
	entities.append(Entity(pgrect(11, 14, 10, 18), 100, 70, ETYPE_FRIENDLY))
	entities[0].gravity=32
	entities[0].setanimation(1)
	entities[0].health=8

	# sword slash
	entities.append(Entity(pgrect(10000, 10000, 0, 0), 0, 0, ETYPE_FRIENDLY))
	entities[1].alive=False

	# hit flash
	entities.append(Entity(pgrect(10000, 10000, 0, 0), 0, 0, ETYPE_FRIENDLY))
	entities[2].alive=False

	curlevel=1

	print "load level %i"%(curlevel)
	ifile=open("level.txt", "r")
	darkside=False
	curenemy=3
	for line in ifile:
		ix=line.find('x')
		if ix==-1:
			darkside=True
			continue
		iy=line.find('y')
		bid=int(line[:ix])
		bx=int(line[ix+1:iy])
		by=int(line[iy+1:])
		if bid==27: # enemy
			entities.append(Entity(pgrect(8, 8, 14, 24), bx/3, by/3-8, ETYPE_ENEMY))
			entities[curenemy].gravity=32
			entities[curenemy].setanimation(8)
			entities[curenemy].health=5
			werepollies.append(Werepolly(entities[curenemy]))
			curenemy+=1
		else:
			if not darkside:
				blocks1.append(Block(bsheet1, bsheet2, templates[bid].brect, templates[bid].crect, bx/3, by/3, templates[bid].solid, False))
			else:
				blocks2.append(Block(bsheet1, bsheet2, templates[bid].brect, templates[bid].crect, bx/3, by/3, templates[bid].solid, True))
	ifile.close()
	print "level loaded"
	
	# test enemy
#	entities.append(Entity(pgrect(8, 8, 14, 24), 150, -10, ETYPE_ENEMY))
#	entities[3].gravity=32
#	entities[3].setanimation(8)
#	entities[3].health=5
#	werepollies.append(Werepolly(entities[3]))

	level=Level()

	for index in range(5):
		input_keys.append(False)

	print "start"
	gameclock=pygame.time.Clock()
	running=True
	scrollx=0
	scrolly=0
	titlescreen=True
	
	debug_frames=0
	starttime=pygame.time.get_ticks()

	global solids
	global lightness
	
	while running:
		pygame.event.pump()
		for event in pygame.event.get():
			if event.type == QUIT:
				running=False

		gameclock.tick(60)

		# input
		for index in range(5):
			input_keys[index]=False
		keys=pygame.key.get_pressed()
		if keys[K_RIGHT]:
			if lightness:
				input_keys[0]=True
			else:
				input_keys[1]=True
		if keys[K_LEFT]:
			if lightness:
				input_keys[1]=True
			else:
				input_keys[0]=True
		if keys[K_z] or keys[K_UP]:
			input_keys[2]=True
		if keys[K_x]:
			input_keys[3]=True
		if keys[K_DOWN]:
			input_keys[4]=True

		if lightness:
			solids=solids1
		else:
			solids=solids2
		
		# update
		if titlescreen:
			if keys[K_z]:
				titlescreen=False
				entities[EID_PLAYER].flag1=True # prevent immediate jump
#				reset_game()
		else:
			update_player()
			update_hitflash()
			for specimen in werepollies:
				if specimen.entity.alive:
					update_werepolly(specimen)
			for entity in entities:
				if entity.alive:
					entity.update()
	
			if entities[EID_PLAYER].x>scrollx/3+90+20:
				scrollx=(entities[EID_PLAYER].x-(90+20))*3
			if entities[EID_PLAYER].x<scrollx/3+90-20:
				scrollx=(entities[EID_PLAYER].x-(90-20))*3

		# draw
		if titlescreen:
			screen.blit(title, (0, 0))
		else:
			debug_frames+=1
			drawn=True
			if lightness: # light
				screen.fill([255,255,255])
				for block in blocks1:
					screen.blit(block.image1, (block.x-scrollx, block.y-scrolly))
				for entity in entities:
					if entity.alive:
						if not entity.flip:
							screen.blit(entity.getsprite(), (entity.x*3-scrollx, entity.y*3-scrolly))
						else:
							screen.blit(pygame.transform.flip(entity.getsprite(), True, False), (entity.x*3-scrollx, entity.y*3-scrolly))
			else: # dark
				screen.fill([0,0,0])
				for block in blocks2:
					screen.blit(pygame.transform.flip(block.image2, True, False), (639-block.x-block.image2.get_width()+scrollx, block.y-scrolly))
				for entity in entities:
					if entity.alive:
						if entity.flip:
							screen.blit(entity.getsprite(), (639-entity.x*3-entity.getsprite().get_width()+scrollx, entity.y*3-scrolly))
						else:
							screen.blit(pygame.transform.flip(entity.getsprite(), True, False), (639-entity.x*3-entity.getsprite().get_width()+scrollx, entity.y*3-scrolly))
			# hud
			for x in range(0,entities[EID_PLAYER].health):
				screen.blit(hud_health1, (8+x*18, 8))
			for x in range(entities[EID_PLAYER].health, 8):
				screen.blit(hud_health2, (8+x*18, 8))
		
		pygame.display.update()
		pygame.display.flip()
	totaltime=pygame.time.get_ticks()-starttime
	print "end: %i frames/sec" % (debug_frames*1000/totaltime)


# ?
if __name__ == "__main__":
	main()

